(* 
   Un insieme finito e' un tipo di dato strutturato, 'a set, della forma {x1,...,xn}, per 
   n>=0, a componenti xi omogenei e contenente al piu'un'occorrenza di uno stesso valore, 
   xi=xj iff i=j. 
   I valori 'a set sono valori Immutable, sui quali possiamo operare mediante le sole se-
   guenti operazioni nell'API SET.
   Si dia un ADT, con nome Set1. Allo scopo, si fornisca:
   (1) Stato della Rappresentazione, AF ed I: Non si sovraccarichi la rappresentazione con
       duplicati di elementi;
   (2) Definizione delle operazioni espressamente indicate;
   (3) Esecuzione e verifica comportamento
 *)

module type SET =
 sig type 'a set 
  val mk: unit -> 'a set
  exception InvalidArgException of string
  val addAll: 'a set -> 'a list -> 'a set
  (* public additional *)
  val toString: 'a set -> ('a -> string) -> string
end;;


(* SOLUZIONE: punto 1 *)
module Set1 =
(struct 
   type 'a set = Empty | Add of 'a * 'a set
   (*
    AF(c) = {} if c = Empty
    AF(c) = {x1,y1,...,yn} if (c = Add(x1,ys)) && (AF(ys) = {y1,...,yn}, n>=0)    
    I(c) = (c = Add(x1,ys)) => not(x1\in AF(ys))
   *)
(* SOLUZIONE: punto 2 *)
   exception InvalidArgException of string
   let mk() = Empty 
   let rec is x set = (match set with
   		|Empty -> false
   		|Add(e,rest) when x = e -> true
   		|Add(_,rest) -> is x rest
   )
   let rec addAll set = (function
   		|[] -> set
   		|x::xs when (is x set) -> addAll set xs
   		|x::xs -> addAll (Add (x,set)) xs
   )
 (* additionals *)
   let rec inner set print = (match set with
   		|Empty -> ""
   		|Add(e,rest) -> (let stringR = inner rest print in 
   						 let stringE = print e in
   						 if stringR = "" then stringE else stringE^", "^stringR)
   )
   let toString set print = "{"^(inner set print)^"}"
end:SET);;


(* SOLUZIONE: punto 3 *)
open Printf;;
let list1 = [1;4;2;6;4;3];;
let s1 = Set1.addAll (Set1.mk()) list1;;
Set1.toString s1 (fun x -> sprintf "%d" x);;
let list2 = [5;4;3;7];;
let s2 = Set1.addAll s1 list2;;
Set1.toString s1 (fun x -> sprintf "%d" x);;


(* 
val list1 : int list = [1; 4; 2; 6; 4; 3]
val s1 : int Set1.set = <abstr>
- : string = "{3, 6, 2, 4, 1}"
val list2 : int list = [5; 4; 3; 7]
val s2 : int Set1.set = <abstr>
- : string = "{3, 6, 2, 4, 1}"  
*)